(**************************************************************************)
(*  Lack - Package manager for non-root users                             *)
(*  (C) 2008 Olivier Schwander <olivier.schwander@ens-lyon.org>           *)
(*  http://chadok.info/lack                                               *)
(*                                                                        *)
(*  This file is part of Lack.                                            *)
(*                                                                        *)
(*  Lack is free software: you can redistribute it and/or modify          *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  Lack is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with Lack.  If not, see <http://www.gnu.org/licenses/>.       *)
(*                                                                        *)
(**************************************************************************)

open Printf
open Filename

open Package
open Utils

let base =
  try Sys.getenv "LACK" with
      Not_found -> (printf "Error: $LACK is not defined\nRun \"eval $(lack-init)\"\n"; exit(1))

let date = chomp (slurp "date +%F")

let rec dependencies quiet verbose needs =
  let needs = List.map (fun u -> (u, Package.parse verbose u base)) needs in
  let needed = List.fold_left (fun acc (u, p) ->
				 if Needs.test verbose base p then
				   acc
				 else
				   (u, p)::acc
			      ) [] needs in
    List.iter (fun (u, p) -> install quiet verbose p u None) needed;
    true

and install quiet verbose pkg source localfile =
  let temp = temp_dir () in
    try
      let name = pkg.pkg_name in
      let version = substitute (total [ ("date", date) ]) pkg.pkg_version in

      let url = pkg.pkg_url in
      let filename = pkg.pkg_file in
      let directory = pkg.pkg_directory in
      let fetch_cmd = pkg.pkg_fetch_cmd in
      let extract_cmd = pkg.pkg_extract_cmd in
      let digest = pkg.pkg_digest in

      let patches = pkg.pkg_patches in

      let needs = pkg.pkg_needs in

      let conf_options = pkg.pkg_conf_options in
      let conf_cmd = pkg.pkg_conf_cmd in

      let build_options = pkg.pkg_build_options in
      let build_cmd = pkg.pkg_build_cmd in

      let install_options = pkg.pkg_install_options in
      let install_cmd = pkg.pkg_install_cmd in

      let paths = pkg.pkg_paths in

      let file =
        match localfile with
          | None -> (match filename with
                       | None -> basename url
                       | Some s -> s)
          | Some s -> basename s
      in
      let target = Steps.prepare verbose temp base name version fetch_cmd url in

      let n = 11 in
      let i = ref 1 in
        Log.task_begin name quiet !i n ("Managing dependencies...");
        Log.status quiet (dependencies quiet verbose needs);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Fetching "^url^"...");
        Log.status quiet (Steps.fetch verbose fetch_cmd url localfile (Some target));
        i := !i + 1;

        Log.task_begin name quiet !i n ("Verifying digest...");
        Log.status quiet (Steps.check verbose target url digest file);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Extracting...");
        Log.status quiet (Steps.extract verbose extract_cmd directory file);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Patching...");
        Log.status quiet (Steps.patch verbose patches);
        i := !i + 1;

        (* We make the difference between configure step and others *)
        Log.task_begin name quiet !i n ("Configuring...");
        Log.status quiet (Steps.run ~configure:true verbose conf_cmd conf_options base name version);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Building...");
        Log.status quiet (Steps.run verbose build_cmd build_options base name version);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Installing...");
        Log.status quiet (Steps.run verbose install_cmd install_options base name version);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Registering...");
        Log.status quiet (Steps.register verbose base name version source paths);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Cleaning build directory...");
        Log.status quiet (Steps.clean verbose base temp);
        i := !i + 1;

        Log.task_begin name quiet !i n ("Updating packages base...");
        Log.status quiet (Init.update base);
        i := !i + 1;

        ()
    with
      | Log.Failure -> printf "Failure !!! Ouch in %s\n" temp
