(**************************************************************************)
(*  Lack - Package manager for non-root users                             *)
(*  (C) 2008 Olivier Schwander <olivier.schwander@ens-lyon.org>           *)
(*  http://chadok.info/lack                                               *)
(*                                                                        *)
(*  This file is part of Lack.                                            *)
(*                                                                        *)
(*  Lack is free software: you can redistribute it and/or modify          *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  Lack is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with Lack.  If not, see <http://www.gnu.org/licenses/>.       *)
(*                                                                        *)
(**************************************************************************)

open Parse
open Parameter
open Utils

exception Invalid_file of string
exception File_not_found of string
exception Invalid_url
exception Invalid_test
exception Missing of string

type version =
  | Exact of string
  | Minmax of string * string
  | List of string list

type dep_dsc = {
  dep_name: string;
  dep_altnames: (string * string) list;
  dep_version: version;
  dep_mandatory: bool;
  yes_imply: parameter list;
  no_imply: parameter list;
  required_paramaters: parameter list;
}

type dependency =
  | Pkg of dep_dsc
  | Or of dependency list

type package = {
  pkg_name : string;
  pkg_version: string;
  pkg_homepage: string option;
  pkg_licence: string option;
  pkg_maintainer: string option;
  pkg_maintainer_email: string option;
  pkg_url: string;
  pkg_file: string option;
  pkg_directory: string;

  pkg_fetch_cmd: string;
  pkg_extract_cmd: string;
  pkg_digest: string option;
  pkg_patches: string list;
  pkg_depends: dependency list;
  pkg_needs: string list;
  pkg_test: string * string;

  pkg_conf_options: parameter list;
  pkg_conf_cmd: string;
  pkg_conf_prefix: parameter option;

  pkg_build_options: parameter list;
  pkg_build_cmd: string;
  pkg_build_prefix: parameter option;

  pkg_install_options: parameter list;
  pkg_install_cmd: string;
  pkg_install_prefix: parameter option;

  pkg_paths: (string * string) list;
}

(* return a tuple with the maintainer name and email or empty strings if
   it is missing *)
let rec get_maintainer l =
  match l with
    | [] -> "", ""
    | x::q ->
        match x with
          | Xml.Element("maintainer", _, l) ->
              find_value l "name", find_value l "email"
          | _ -> get_maintainer q

(* return the list of patches url *)
let get_patches l =
  let p = find_node l "patches" in
  let f acc x =
    match x with
      | Xml.Element("url", _, [Xml.PCData(url)]) ->
          url::acc
      | _ -> acc
  in
    List.rev (Xml.fold f [] p)

(* return the list of dependencies url *)
let get_needs l =
  try
    let p = find_node l "needs" in
    let f acc x =
      match x with
        | Xml.Element("url", _, [Xml.PCData(url)]) ->
            url::acc
        | _ -> acc
    in
      List.rev (Xml.fold f [] p)
  with
      Not_found -> []

(* returns the paths node if existing *)
let get_paths l =
  let p = find_node l "paths" in
  let f acc x =
    match x with
      | Xml.Element("path", _, [Xml.PCData(path)]) ->
          let name = Xml.attrib x "name" in
            (name, path)::acc
      | _ -> acc
  in
    List.rev (Xml.fold f [] p)

(* get the test to know if the package is present *)
let get_test l =
  try
    let t = find_node l "test" in
    let value = match t with
      | Xml.Element("test", _, [Xml.PCData(value)]) -> value
      | _ -> raise Invalid_test
    in
    let kind = try Xml.attrib t "kind" with
      | Xml.No_attribute _ -> "exec"
    in
      value, kind
  with
      Not_found -> Printf.fprintf stderr "Warning: field 'test' is missing\n"; ("true", "exec")

(* return the url of the sources, maybe with specific values for
   fetching and extracting or default values *)
let get_url l version =
  let url = find_node l "url" in
  let url_value =
    match url with
      | Xml.Element("url", _, [Xml.PCData(value)]) -> value
      | _ -> raise Invalid_url
  in
    (* search for an attribute, or take the default value *)
  let defaults s = try Xml.attrib url s with
    | Xml.No_attribute _ -> Defaults.give s in
  let url_value = substitute (total [ ("version", version) ]) url_value in
  let fetch_cmd =  defaults "fetch" in
  let extract_cmd = defaults "extract" in
  let file = try Some (Xml.attrib url "filename") with
    | Xml.No_attribute _ -> None
  in
    url_value, fetch_cmd, extract_cmd, file

(* used for the three main steps: configure, build, install. Get the
   command name, its parameters and a possible prefix field *)
let get_instr step l =
  let conf = find_node l step in
  let command = find_value (Xml.children conf) "command" in
    (* get all the parameters *)
  let options =
    try
      let opts = find_node (Xml.children conf) "options" in
      let f acc x =
        match x with
          | Xml.Element("arg", _, [Xml.PCData(s)]) ->
              (Arg s)::acc
          | Xml.Element("env", _, [Xml.PCData(s)]) ->
              let name = Xml.attrib x "name" in
                (Env (name,s))::acc
          | _ -> acc
      in
        List.rev (Xml.fold f [] opts)
    with
        Not_found -> []
  in
    (* get the prefix *)
  let prefix =
    try
      let node = find_node (Xml.children conf) "prefix" in
        match List.hd (Xml.children node) with
          | Xml.Element("arg", _, [Xml.PCData(s)]) -> Some (Arg s)
          | Xml.Element("env", _, [Xml.PCData(s)]) ->
              let name = Xml.attrib node "name" in
                Some (Env (name,s))
          | _ -> raise (Invalid_argument "Missing parameter for prefix")
    with
        (* a prefix is mandatory for the configuation step, so use the
           default value if there is nothing in the package description *)
      | Not_found ->
          if step = "configuration" then
            Some (Arg (Defaults.give "prefix"))
          else
            None
  in
    command, options, prefix

(* convert an empty string to None *)
let none_if_empty s =
  if s = "" then
    None
  else
    Some s

(* change an empty string to a default value *)
let default_if_empty s =
  if s = "" then
    Defaults.give s
  else
    s

(* failure if a mandatory value is missing *)
let non_empty name value =
  if value = "" then
    raise (Missing name)
  else
    value

let retrieve verbose url base =
  let may_http = String.sub url 0 7 in
  let may_https = String.sub url 0 8 in
  let may_ftp = String.sub url 0 6 in
  let filename = Filename.basename url in
    chdir base;
    (if may_http = "http://" or may_https = "https://" or may_ftp = "ftp://" then
      ignore (command verbose ("wget -N "^url))
    else
      ignore (command verbose ("cp "^url^" .")));
    filename

let parse verbose url base =
  let file = retrieve verbose url base in
  let x = try Xml.parse_file file with
    | Xml.File_not_found _ -> raise (File_not_found file) in
    match x with
      | Xml.Element("package", _, l) ->
          let name       = find_value l "name" in
          let version    = find_value l "version" in
          let licence    = find_value l "licence" in
          let homepage   = find_value l "homepage" in
          let maintainer = get_maintainer l in
          let url, fetch, extract, file = get_url l version in
          let directory  = find_value l "directory" in
          let digest     = find_value l "digest" in
          let patches    = get_patches l in
          let needs      = get_needs l in
          let test       = get_test l in
          let conf_cmd, conf_options, conf_prefix = get_instr "configuration" l in
          let build_cmd, build_options, build_prefix = get_instr "build" l in
          let install_cmd, install_options, install_prefix = get_instr "install" l in
          let paths = get_paths l in
            {
              pkg_name = non_empty "name" name;
              pkg_version = non_empty "version" version;

              pkg_licence = none_if_empty licence;
              pkg_homepage = none_if_empty homepage;
              pkg_maintainer = none_if_empty (fst maintainer);
              pkg_maintainer_email = none_if_empty (snd maintainer);

              pkg_url = non_empty "url" url;
              pkg_directory = non_empty "directory" directory;
              pkg_fetch_cmd = fetch;
              pkg_file = file;
              pkg_extract_cmd = extract;

              pkg_digest = none_if_empty digest;

              pkg_patches = patches;

              pkg_depends = [];
              pkg_needs = needs;
              pkg_test = test;

              pkg_conf_options = conf_options;
              pkg_conf_cmd = conf_cmd;
              pkg_conf_prefix = conf_prefix;

              pkg_build_options = build_options;
              pkg_build_cmd = build_cmd;
              pkg_build_prefix = build_prefix;

              pkg_install_options = install_options;
              pkg_install_cmd = install_cmd;
              pkg_install_prefix = install_prefix;

              pkg_paths = paths;
            }
      | _ -> raise (Invalid_file file)

