(**************************************************************************)
(*  Lack - Package manager for non-root users                             *)
(*  (C) 2008 Olivier Schwander <olivier.schwander@ens-lyon.org>           *)
(*  http://chadok.info/lack                                               *)
(*                                                                        *)
(*  This file is part of Lack.                                            *)
(*                                                                        *)
(*  Lack is free software: you can redistribute it and/or modify          *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  Lack is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with Lack.  If not, see <http://www.gnu.org/licenses/>.       *)
(*                                                                        *)
(**************************************************************************)

open Printf
open Filename

open Parameter
open Utils

exception Invalid_digest
exception Patch_error of string

let prepare verbose temp prefix_path name version cmd url =
  let target = temp in
    (try mkdir target with
       | _ -> ()
    );
    (try mkdir (prefix_path^"/"^name) with
       | _ -> ()
    );
    (try mkdir (prefix_path^"/"^name^"/"^version) with
       | _ -> ()
    );
    target^"/"

let fetch verbose cmd url file target =
  (match target with
    | None -> ()
    | Some d -> chdir d);
  match file with
    | None -> command verbose (cmd^" '"^url^"'")
    | Some f -> command verbose ("cp -r "^f^" .")

let check verbose target url digest file =
  match digest with
    | None -> true
    | Some d ->
        let fresh_digest = Digest.to_hex (Digest.file (target^file)) in
          if d = fresh_digest then
            true
          else
            raise Invalid_digest

let extract verbose extract directory file =
  let st = command verbose (extract^" "^file) in
    chdir directory;
    st

let patch verbose patch_list =
  let apply url =
    let file = basename url in
      if fetch verbose "wget" url None None then
        ()
      else
        raise (Patch_error url);
      let file =
        if check_suffix file ".gz" then (
          if command verbose ("gunzip "^file) then
            ()
          else
            raise (Patch_error url);
          chop_extension file
        )
        else
          file
      in
        if verbose then
          printf "Applying %s\n%!" file;
        if command verbose ("patch -p1 <"^file) then
          ()
        else
          raise (Patch_error url);
  in
    try (List.iter apply patch_list; true) with _ -> false

let run ?(configure=false) verbose cmd options prefix_path name version =
  let prefix = build_prefix prefix_path name version in
  let arg_line = build_args ~configure:configure prefix options in
  let cmd = substitute (total ["PREFIX", prefix;]) cmd in
    if cmd <> "" then
      let cmd = if arg_line <> "" then cmd^" "^arg_line else cmd in
        if verbose then
          print_endline cmd;
        command verbose cmd
    else
      true

let register verbose prefix name version source paths =
  let gen_node name =
    fun value -> Xml.Element(name, [], [Xml.PCData value])
  in
  let gen_path name =
    fun value -> Xml.Element("path", ["name", name], [Xml.PCData value])
  in
  let install_path = prefix^"/"^name^"/"^version in
  let current_path = prefix^"/"^name^"/"^"/current" in
  let find = Find.find ~verbose:verbose install_path in
  let build dir =
    let l = find dir in
      List.map (gen_path dir) l
  in
    try
      ignore (command verbose ("rm -f "^current_path));
      ignore (command verbose ("ln -s "^install_path^" "^current_path));
      let children = 
        match paths with
          | [] ->
              List.flatten (List.map build Meta.sources)
          | l -> List.map (fun (n, p) -> gen_path n (Filename.concat install_path p)) l
      in
      let meta = Xml.Element("meta", [], [
                               gen_node "name" name;
                               gen_node "version" version;
                               gen_node "source" source;
                               Xml.Element("paths", [], children);
                             ])
        
      in
      let file = open_out (install_path^"/meta.xml") in
        output_string file ((Xml.to_string_fmt meta)^"\n");
        close_out file;
        true
    with
        _ -> false

let clean verbose base dir =
  chdir base;
  command verbose ("rm -fr "^dir);

